"""Tests that require external access (e.g. pip install, virtualenv creation)"""
import os
import subprocess
import sys

import pytest

if sys.version_info[:2] >= (3, 4):
    from pathlib import Path
else:
    from pathlib2 import Path

from tests.lib import need_git


@pytest.mark.network
def test_package_setuptools(initproj, cmd):
    initproj(
        "magic-0.1",
        filedefs={
            "tox.ini": """\
                [tox]
                isolated_build = true
                [testenv:.package]
                basepython = {}
            """.format(
                sys.executable,
            ),
            "pyproject.toml": """\
                [build-system]
                requires = ["setuptools >= 35.0.2", "setuptools_scm >= 2.0.0, <3"]
                build-backend = "setuptools.build_meta"
            """,
        },
    )
    run(cmd, "magic-0.1.tar.gz")


@pytest.mark.network
@need_git
@pytest.mark.skipif(sys.version_info < (3, 0), reason="flit is Python 3 only")
def test_package_flit(initproj, cmd):
    initproj(
        "magic-0.1",
        filedefs={
            "tox.ini": """\
                [tox]
                isolated_build = true
                [testenv:.package]
                basepython = {}
            """.format(
                sys.executable,
            ),
            "pyproject.toml": """\
                [build-system]
                requires = ["flit"]
                build-backend = "flit.buildapi"

                [tool.flit.metadata]
                module = "magic"
                author = "Happy Harry"
                author-email = "happy@harry.com"
                home-page = "https://github.com/happy-harry/is"
                requires = [
                    "tox",
                ]
            """,
            ".gitignore": ".tox",
        },
        add_missing_setup_py=False,
    )
    env = os.environ.copy()
    env["GIT_COMMITTER_NAME"] = "committer joe"
    env["GIT_AUTHOR_NAME"] = "author joe"
    env["EMAIL"] = "joe@example.com"
    subprocess.check_call(["git", "init"], env=env)
    subprocess.check_call(["git", "add", "-A", "."], env=env)
    subprocess.check_call(["git", "commit", "-m", "first commit", "--no-gpg-sign"], env=env)

    run(cmd, "magic-0.1.tar.gz")


@pytest.mark.network
@pytest.mark.skipif(sys.version_info < (3, 0), reason="poetry is Python 3 only")
def test_package_poetry(initproj, cmd):
    initproj(
        "magic-0.1",
        filedefs={
            "tox.ini": """\
                [tox]
                isolated_build = true
                [testenv:.package]
                basepython = {}
            """.format(
                sys.executable,
            ),
            "pyproject.toml": """\
                [build-system]
                requires = ["poetry>=0.12"]
                build-backend = "poetry.masonry.api"

                [tool.poetry]
                name = "magic"
                version = "0.1.0"
                description = ""
                authors = ["Name <email@email.com>"]
            """,
            ".gitignore": ".tox",
        },
        add_missing_setup_py=False,
    )
    run(cmd, "magic-0.1.0.tar.gz")


def run(cmd, package):
    result = cmd("--sdistonly", "-e", "py", "-v", "-v")
    result.assert_success(is_run_test_env=False)
    package_venv = (Path() / ".tox" / ".package").resolve()
    assert ".package create: {}".format(package_venv) in result.outlines, result.out
    assert "write config to {}".format(package_venv / ".tox-config1") in result.out, result.out
    package_path = (Path() / ".tox" / "dist" / package).resolve()
    assert package_path.exists()

    package_path.unlink()

    # second call re-uses
    result2 = cmd("--sdistonly", "-e", "py", "-v", "-v")

    result2.assert_success(is_run_test_env=False)
    assert (
        ".package reusing: {}".format(package_venv) in result2.outlines
    ), "Second call output:\n{}First call output:\n{}".format(result2.out, result.out)
    assert package_path.exists()



import os
import traceback

import py
from flaky import flaky

from tox.session.commands.run import sequential


@flaky(max_runs=3)
def test_tox_parallel_build_safe(initproj, cmd, mock_venv, monkeypatch):
    initproj(
        "env_var_test",
        filedefs={
            "tox.ini": """
                  [tox]
                  envlist = py
                  install_cmd = python -m -c 'print("ok")' -- {opts} {packages}'
                  [testenv]
                  commands = python -c 'import sys; print(sys.version)'
                      """,
        },
    )
    # we try to recreate the following situation
    # t1 starts and performs build
    # t2 starts, but is blocked from t1 build lock to build
    # t1 gets unblocked, t2 can now enter
    # t1 is artificially blocked to run test command until t2 finishes build
    #  (parallel build package present)
    # t2 package build finishes both t1 and t2 can now finish and clean up their build packages
    import threading

    import tox.package

    t1_build_started = threading.Event()
    t1_build_blocker = threading.Event()
    t2_build_started = threading.Event()
    t2_build_finished = threading.Event()

    invoke_result = {}

    def invoke_tox_in_thread(thread_name):
        try:
            result = cmd("--parallel--safe-build", "-vv")
        except Exception as exception:
            result = exception, traceback.format_exc()
        invoke_result[thread_name] = result

    prev_build_package = tox.package.build_package

    with monkeypatch.context() as m:

        def build_package(config, session):
            t1_build_started.set()
            t1_build_blocker.wait()
            return prev_build_package(config, session)

        m.setattr(tox.package, "build_package", build_package)

        prev_run_test_env = sequential.runtestenv

        def run_test_env(venv, redirect=False):
            t2_build_finished.wait()
            return prev_run_test_env(venv, redirect)

        m.setattr(sequential, "runtestenv", run_test_env)

        t1 = threading.Thread(target=invoke_tox_in_thread, args=("t1",))
        t1.start()
        t1_build_started.wait()

    with monkeypatch.context() as m:

        def build_package(config, session):
            t2_build_started.set()
            try:
                return prev_build_package(config, session)
            finally:
                t2_build_finished.set()

        m.setattr(tox.package, "build_package", build_package)

        t2 = threading.Thread(target=invoke_tox_in_thread, args=("t2",))
        t2.start()

        # t2 should get blocked by t1 build lock
        t2_build_started.wait(timeout=0.1)
        assert not t2_build_started.is_set()

        t1_build_blocker.set()  # release t1 blocker -> t1 can now finish
        # t1 at this point should block at run test until t2 build finishes
        t2_build_started.wait()

    t1.join()  # wait for both t1 and t2 to finish
    t2.join()

    # all threads finished without error
    for val in invoke_result.values():
        if isinstance(val, tuple):
            assert False, "{!r}\n{}".format(val[0], val[1])
    err = "\n".join(
        "{}=\n{}".format(k, v.err).strip() for k, v in invoke_result.items() if v.err.strip()
    )
    out = "\n".join(
        "{}=\n{}".format(k, v.out).strip() for k, v in invoke_result.items() if v.out.strip()
    )
    for val in invoke_result.values():
        assert not val.ret, "{}\n{}".format(err, out)
    assert not err

    # when the lock is hit we notify
    lock_file = py.path.local().join(".tox", ".package.lock")
    msg = "lock file {} present, will block until released".format(lock_file)
    assert msg in out

    # intermediate packages are removed at end of build
    t1_package = invoke_result["t1"].session.getvenv("py").package
    t2_package = invoke_result["t1"].session.getvenv("py").package
    assert t1 != t2
    assert not t1_package.exists()
    assert not t2_package.exists()

    # the final distribution remains
    dist_after = invoke_result["t1"].session.config.distdir.listdir()
    assert len(dist_after) == 1
    sdist = dist_after[0]
    assert t1_package != sdist
    # our set_os_env_var is not thread-safe so clean-up TOX_WORK_DIR
    os.environ.pop("TOX_WORK_DIR", None)



import shutil

from tox import reporter


def ensure_empty_dir(path):
    if path.check():
        reporter.info("  removing {}".format(path))
        shutil.rmtree(str(path), ignore_errors=True)
        path.ensure(dir=1)



def test_listenvs(cmd, initproj, monkeypatch):
    monkeypatch.delenv(str("TOXENV"), raising=False)
    initproj(
        "listenvs",
        filedefs={
            "tox.ini": """
        [tox]
        envlist=py36,py27,py37,pypi,docs
        description= py27: run pytest on Python 2.7
                     py37: run pytest on Python 3.6
                     pypi: publish to PyPI
                     docs: document stuff
                     notincluded: random extra

        [testenv:notincluded]
        changedir = whatever

        [testenv:docs]
        changedir = docs
        """,
        },
    )

    result = cmd("-l")
    assert result.outlines == ["py36", "py27", "py37", "pypi", "docs"]

    result = cmd("-l", "-e", "py")
    assert result.outlines == ["py36", "py27", "py37", "pypi", "docs"]

    monkeypatch.setenv(str("TOXENV"), str("py"))
    result = cmd("-l")
    assert result.outlines == ["py36", "py27", "py37", "pypi", "docs"]

    monkeypatch.setenv(str("TOXENV"), str("py36"))
    result = cmd("-l")
    assert result.outlines == ["py36", "py27", "py37", "pypi", "docs"]


def test_listenvs_verbose_description(cmd, initproj):
    initproj(
        "listenvs_verbose_description",
        filedefs={
            "tox.ini": """
        [tox]
        envlist=py36,py27,py37,pypi,docs
        [testenv]
        description= py36: run pytest on Python 3.6
                     py27: run pytest on Python 2.7
                     py37: run pytest on Python 3.7
                     pypi: publish to PyPI
                     docs: document stuff
                     notincluded: random extra

        [testenv:notincluded]
        changedir = whatever

        [testenv:docs]
        changedir = docs
        description = let me overwrite that
        """,
        },
    )
    result = cmd("-lv")
    expected = [
        "default environments:",
        "py36 -> run pytest on Python 3.6",
        "py27 -> run pytest on Python 2.7",
        "py37 -> run pytest on Python 3.7",
        "pypi -> publish to PyPI",
        "docs -> let me overwrite that",
    ]
    assert result.outlines[2:] == expected


def test_listenvs_all(cmd, initproj, monkeypatch):
    initproj(
        "listenvs_all",
        filedefs={
            "tox.ini": """
        [tox]
        envlist=py36,py27,py37,pypi,docs

        [testenv:notincluded]
        changedir = whatever

        [testenv:docs]
        changedir = docs
        """,
        },
    )
    result = cmd("-a")
    expected = ["py36", "py27", "py37", "pypi", "docs", "notincluded"]
    assert result.outlines == expected

    result = cmd("-a", "-e", "py")
    assert result.outlines == ["py36", "py27", "py37", "pypi", "docs", "py", "notincluded"]

    monkeypatch.setenv(str("TOXENV"), str("py"))
    result = cmd("-a")
    assert result.outlines == ["py36", "py27", "py37", "pypi", "docs", "py", "notincluded"]

    monkeypatch.setenv(str("TOXENV"), str("py36"))
    result = cmd("-a")
    assert result.outlines == ["py36", "py27", "py37", "pypi", "docs", "notincluded"]


def test_listenvs_all_verbose_description(cmd, initproj):
    initproj(
        "listenvs_all_verbose_description",
        filedefs={
            "tox.ini": """
        [tox]
        envlist={py27,py36}-{windows,linux} # py35
        [testenv]
        description= py27: run pytest on Python 2.7
                     py36: run pytest on Python 3.6
                     windows: on Windows platform
                     linux: on Linux platform
                     docs: generate documentation
        commands=pytest {posargs}

        [testenv:docs]
        changedir = docs
        """,
        },
    )
    result = cmd("-av")
    expected = [
        "default environments:",
        "py27-windows -> run pytest on Python 2.7 on Windows platform",
        "py27-linux   -> run pytest on Python 2.7 on Linux platform",
        "py36-windows -> run pytest on Python 3.6 on Windows platform",
        "py36-linux   -> run pytest on Python 3.6 on Linux platform",
        "",
        "additional environments:",
        "docs         -> generate documentation",
    ]
    assert result.outlines[-len(expected) :] == expected


def test_listenvs_all_verbose_description_no_additional_environments(cmd, initproj):
    initproj(
        "listenvs_all_verbose_description",
        filedefs={
            "tox.ini": """
        [tox]
        envlist=py27,py36
        """,
        },
    )
    result = cmd("-av")
    expected = ["default environments:", "py27 -> [no description]", "py36 -> [no description]"]
    assert result.out.splitlines()[-3:] == expected
    assert "additional environments" not in result.out


def test_listenvs_packaging_excluded(cmd, initproj):
    initproj(
        "listenvs",
        filedefs={
            "tox.ini": """
        [tox]
        envlist = py36,py27,py37,pypi,docs
        isolated_build = True

        [testenv:notincluded]
        changedir = whatever

        [testenv:docs]
        changedir = docs
        """,
        },
    )
    result = cmd("-a")
    expected = ["py36", "py27", "py37", "pypi", "docs", "notincluded"]
    assert result.outlines == expected, result.outlines


def test_listenvs_all_extra_definition_order_decreasing(cmd, initproj):
    initproj(
        "listenvs_all",
        filedefs={
            "tox.ini": """
        [tox]
        envlist=py36

        [testenv:b]
        changedir = whatever

        [testenv:a]
        changedir = docs
        """,
        },
    )
    result = cmd("-a")
    expected = ["py36", "b", "a"]
    assert result.outlines == expected


def test_listenvs_all_extra_definition_order_increasing(cmd, initproj):
    initproj(
        "listenvs_all",
        filedefs={
            "tox.ini": """
        [tox]
        envlist=py36

        [testenv:a]
        changedir = whatever

        [testenv:b]
        changedir = docs
        """,
        },
    )
    result = cmd("-a")
    expected = ["py36", "a", "b"]
    assert result.outlines == expected


def test_listenvs_without_default_envs(cmd, initproj):
    """When running tox -l without any default envirinments, nothing happens."""
    initproj(
        "logsnada",
        filedefs={"tox.ini": ""},
    )
    result = cmd("-l")
    assert result.ret == 0
    assert result.out == ""



# -*- coding: utf-8 -*-
"""Handles creating a release PR"""
import base64
import json
import os
import tempfile
import textwrap
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path
from typing import Tuple

import httplib2
from apiclient import discovery
from git import Remote, Repo
from oauth2client import client, file, tools
from packaging.version import Version

ROOT_SRC_DIR = Path(__file__).parents[1]


def main() -> None:
    repo = Repo(str(ROOT_SRC_DIR))
    update_upstream(repo)
    prev_version, release_version = get_last_release_versions(repo)
    send_mail_message(
        subject=f"tox release {release_version}",
        content=get_message_body(release_version, prev_version),
    )
    print("All done! ✨ 🍰 ✨")


def get_message_body(release_version: Version, prev_version: Version) -> str:
    is_major_release = release_version.release[0:2] != prev_version.release[0:2]
    if is_major_release:
        return textwrap.dedent(
            f"""
        The tox team is proud to announce the {release_version} feature release at https://pypi.org/project/tox/!

        tox aims to automate and standardize testing in Python. It is part of a larger vision of easing the packaging, testing and release process of Python software.

        Details about the changes can be found at https://tox.readthedocs.io/en/{release_version}/changelog.html
        For complete documentation, please visit: https://tox.readthedocs.io/en/{release_version}/

        As usual, you can upgrade from PyPI via:

            pip install --upgrade tox

        or - if you also want to get pre release versions:

            pip install -upgrade --pre tox

        We thank all present and past contributors to tox. Have a look at https://github.com/tox-dev/tox/blob/master/CONTRIBUTORS to see who contributed.

        Happy toxing,
        the tox-dev team
        """,  # noqa
        )
    else:
        return textwrap.dedent(
            f"""
                The tox team is proud to announce the {release_version} bug fix release at https://pypi.org/project/tox/!

                tox aims to automate and standardize testing in Python. It is part of a larger vision of easing the packaging, testing and release process of Python software.

                For details about the fix(es),please check the CHANGELOG: https://tox.readthedocs.io/en/{release_version}/changelog.html

                We thank all present and past contributors to tox. Have a look at https://github.com/tox-dev/tox/blob/master/CONTRIBUTORS to see who contributed.

                Happy toxing,
                the tox-dev team
                """,  # noqa
        )


def get_upstream(repo: Repo) -> Remote:
    for remote in repo.remotes:
        for url in remote.urls:
            if url.endswith("tox-dev/tox.git"):
                return remote
    raise RuntimeError("could not find tox-dev/tox.git remote")


def get_last_release_versions(repo: Repo) -> Tuple[Version, Version]:
    print("get latest release version")
    commit_to_tag = {tag.commit.hexsha: tag for tag in repo.tags}
    _, release_tag = sorted(
        [(tag.commit.committed_datetime, tag) for tag in repo.tags],
        reverse=True,
    )[0]
    for commit in release_tag.commit.iter_parents():
        if commit.hexsha in commit_to_tag:
            prev_release_tag = commit_to_tag[commit.hexsha]
            prev_version = Version(prev_release_tag.name)
            if not any(
                (
                    prev_version.is_devrelease,
                    prev_version.is_prerelease,
                    prev_version.is_postrelease,
                ),
            ):
                break
    else:
        raise RuntimeError("no previous release")
    release_version = Version(release_tag.name)
    print(f"\trelease {release_version} with previous {prev_version}")
    return prev_version, release_version


def update_upstream(repo: Repo) -> None:
    print("fetch latest remote")
    upstream = get_upstream(repo)
    upstream.fetch()


def send_mail_message(subject, content):
    this_dir = Path(__file__).parent
    store = file.Storage("credentials.json")
    credentials = store.get()
    if not credentials or credentials.invalid:
        client_secret_json = json.loads((this_dir / "client_secret.json").read_text())
        client_secret_json["installed"]["client_secret"] = os.environ["TOX_DEV_GOOGLE_SECRET"]
        with tempfile.NamedTemporaryFile(mode="w+t") as temp_filename:
            json.dump(client_secret_json, temp_filename)
            temp_filename.flush()
            flow = client.flow_from_clientsecrets(
                filename=temp_filename.name,
                scope="https://www.googleapis.com/auth/gmail.send",
            )
            credentials = tools.run_flow(flow, store)
    service = discovery.build("gmail", "v1", http=credentials.authorize(httplib2.Http()))

    message = MIMEMultipart("alternative")
    message["Subject"] = subject
    message["From"] = "toxdevorg@gmail.com"
    recipients = ["testing-in-python@lists.idyll.org", "tox-dev@python.org"]
    message["To"] = ", ".join(recipients)
    message.attach(MIMEText(content, "plain"))
    raw_message_no_attachment = base64.urlsafe_b64encode(message.as_bytes())
    raw_message_no_attachment = raw_message_no_attachment.decode()
    body = {"raw": raw_message_no_attachment}
    message_sent = service.users().messages().send(userId="me", body=body).execute()
    message_id = message_sent["id"]
    print(f"\tMessage sent with id: {message_id}")


if __name__ == "__main__":
    main()



import pytest


def test_parallel_default(newconfig):
    config = newconfig([], "")
    assert isinstance(config.option.parallel, int)
    assert config.option.parallel == 0
    assert config.option.parallel_live is False


def test_parallel_live_on(newconfig):
    config = newconfig(["-o"], "")
    assert config.option.parallel_live is True


def test_parallel_auto(newconfig):
    config = newconfig(["-p", "auto"], "")
    assert isinstance(config.option.parallel, int)
    assert config.option.parallel > 0


def test_parallel_all(newconfig):
    config = newconfig(["-p", "all"], "")
    assert config.option.parallel is None


def test_parallel_number(newconfig):
    config = newconfig(["-p", "2"], "")
    assert config.option.parallel == 2


def test_parallel_number_negative(newconfig, capsys):
    with pytest.raises(SystemExit):
        newconfig(["-p", "-1"], "")

    out, err = capsys.readouterr()
    assert not out
    assert "value must be positive" in err


def test_depends(newconfig, capsys):
    config = newconfig(
        """\
        [tox]
        [testenv:py]
        depends = py37, py36
        """,
    )
    assert config.envconfigs["py"].depends == ("py37", "py36")


def test_depends_multi_row_facotr(newconfig, capsys):
    config = newconfig(
        """\
        [tox]
        [testenv:py]
        depends = py37,
                  {py36}-{a,b}
        """,
    )
    assert config.envconfigs["py"].depends == ("py37", "py36-a", "py36-b")


def test_depends_factor(newconfig, capsys):
    config = newconfig(
        """\
        [tox]
        [testenv:py]
        depends = {py37, py36}-{cov,no}
        """,
    )
    assert config.envconfigs["py"].depends == ("py37-cov", "py37-no", "py36-cov", "py36-no")
